﻿using System;
using System.Collections.Concurrent;
using System.Linq.Expressions;
using System.Threading.Tasks;

namespace ValidateIndex
{
    public class Observable<T> : IObservable<T>
    {
        private readonly ConcurrentDictionary<int, IObserver<T>> _observers = new ConcurrentDictionary<int, IObserver<T>>();

        protected void Notify(T obj)
        {
            foreach (var observer in _observers.Values)
            {
                observer.OnNext(obj);
            }
        }

        public IDisposable Subscribe(IObserver<T> observer)
        {
            if (!_observers.ContainsKey(observer.GetHashCode()))
            {
                _observers.TryAdd(observer.GetHashCode(), observer);
            }
            return new Unsubscriber(_observers, observer);
        }

        private class Unsubscriber : IDisposable
        {
            private readonly ConcurrentDictionary<int, IObserver<T>> _observers;
            private readonly IObserver<T> _observer;

            public Unsubscriber(ConcurrentDictionary<int, IObserver<T>> observers, IObserver<T> observer)
            {
                _observers = observers;
                _observer = observer;
            }

            public void Dispose()
            {
                Task.Factory.StartNew(() =>
                {
                    if (_observer == null || !_observers.ContainsKey(_observer.GetHashCode()))
                        return;
                    IObserver<T> removedObserver;
                    _observers.TryRemove(_observer.GetHashCode(), out removedObserver);
                });
            }
        }
    }

    public class Observable : Observable<string>
    {
        public virtual void RaisePropertyChanged<T>(Expression<Func<T>> prop)
        {
            Notify(Interaction.GetPropertyName(prop));
        }
    }
}
